#!/usr/bin/env python3

import argparse
import asyncio
from pathlib import Path
from functools import partial

import websockets
from .mmelodies_pb2 import UpdateRequest, Update


CONNECTIONS = {}


def connections(*paths):
    if not paths:
        paths = CONNECTIONS.keys()

    clients = []
    for path in paths:
        clients.extend(CONNECTIONS.get(path, []))

    return clients


async def request_handler(ws, writer):
    print(f"connection on {ws.path}")

    if ws.path not in CONNECTIONS.keys():
        CONNECTIONS[ws.path] = []

    CONNECTIONS[ws.path].append(ws)

    while True:
        try:
            message = await ws.recv()

            request = UpdateRequest()
            request.ParseFromString(message)

            if request.scope_samples:
                writer.write(b"scope;")
            if request.param_refresh:
                writer.write(b"params;")
            if request.pgm_event:
                writer.write(f"pgmevt {request.pgm_event};".encode())

            for param in request.param_changes:
                writer.write(f"set {param.id} {param.value};".encode())

            writer.write(b"\n")

            await writer.drain()
        except TypeError as e:
            print(repr(e))
        except websockets.ConnectionClosedOK:
            CONNECTIONS[ws.path].remove(ws)
            break
        except websockets.ConnectionClosedError as e:
            print(e)
            CONNECTIONS[ws.path].remove(ws)
            break


async def update_handler(reader):
    while True:
        message = await reader.readline()

        if not message:
            raise Exception("Received an empty message, connection likely closed.")

        try:
            message = message.decode().strip(";\n")

            update = Update()
            parts = message.split(" ")

            if parts[0] == "params":
                for param_id, value in enumerate(parts[1:]):
                    param = update.params.add()
                    param.id = param_id
                    param.value = int(value)
                    clients = connections()
            elif parts[0] == "scope":
                update.scope_samples.extend([float(part) for part in parts[1:]])
                clients = connections("/api/display")
            elif parts[0] == "change":
                param = update.params.add()
                param.id = int(parts[1])
                param.value = int(parts[2])
                clients = connections()
            else:
                print(parts[0])
                continue

            websockets.broadcast(clients, update.SerializeToString())
        except Exception as e:
            print(repr(e))


async def _main(listen_port, pd_port):
    if pd_port[0] == "unix":
        path = pd_port[1]
        for _ in range(5):
            try:
                print(f"connecting to Pd on {path.absolute()}")
                pd_reader, pd_writer = await asyncio.open_unix_connection(
                    path=path, happy_eyeballs_delay=0.25
                )
            except OSError as e:
                print(e)
                await asyncio.sleep(5)
            else:
                break
        else:
            raise Exception("Could not connect to pd!")

    if pd_port[0] == "tcp":
        host, port = pd_port[1]
        for _ in range(5):
            try:
                print(f"connecting to Pd on {host}:{port}")
                pd_reader, pd_writer = await asyncio.open_connection(
                    host=host, port=port, happy_eyeballs_delay=0.25
                )
            except OSError as e:
                print(e)
                await asyncio.sleep(5)
            else:
                break
        else:
            raise Exception("Could not connect to pd!")

    ws_handler = partial(request_handler, writer=pd_writer)

    if listen_port[0] == "unix":
        path = listen_port[1]
        print(f"listening on {path.absolute()}")
        ws = websockets.unix_serve(ws_handler, path=path)
    if listen_port[0] == "tcp":
        host, port = listen_port[1]
        print(f"listening on {host}:{port}")
        ws = websockets.serve(ws_handler, host=host, port=port)

    async with ws:
        while True:
            await update_handler(pd_reader)


def port(arg: str):
    kind, address = arg.split(":", maxsplit=1)
    if kind == "tcp":
        host, port = address.rsplit(":", maxsplit=1)
        port = int(port)
        if host == "":
            host = "localhost"
        return kind, (host, port)

    if kind == "unix":
        path = Path(address)
        assert path.parent.exists()
        return kind, path

    raise ValueError(f"Unknown port type {kind}")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-l", "--listen-port", type=port, default="tcp:localhost:8001")
    parser.add_argument("-p", "--pd-port", type=port, default="tcp:localhost:10234")
    args = parser.parse_args()

    asyncio.run(_main(**vars(args)))


if __name__ == "__main__":
    main()
